#!/usr/bin/env python3
from io import BytesIO
from pequod_agent.cluster.manager.base import CoreManager, AppNode
from pequod_agent.cluster.service import CoreService, RequestException
from pequod_agent.models.version_range import version_matches
from pequod_agent.models.zone import Zone
from pequod_agent.networking.ip import get_host_ip
from pequod_agent.networking.routing.service import ServiceEndpoint
from time import time
import logging
import tarfile


# TODO separate this in independent services???

class ClusterCore(CoreManager):
    def __init__(self, seed_ips, credentials):
        self.log = logging.getLogger('pequod.core-manager')
        self.seed_ips = seed_ips
        self.credentials = credentials
        self.service = CoreService(seed_ips, 'core-manager', credentials)
        self._last_got_external_services = 0
        self._last_external_services = []

    def get_vaults(self):
        return ClusterVaults(self.seed_ips, self.credentials)

    def get_node(self, name: str) -> dict():
        try:
            data = self.service.get('/nodes/{node_name}'.format(node_name=name))
        except RequestException:
            data = {}
        return AppNode(self, name, data)

    def reset_reboot_flag(self, name: str) -> None:
        data = {'reboot': False, 'maintenance': False}
        self.service.patch('/nodes/{node_name}'.format(node_name=name), data=data)

    def get_networks(self) -> list:
        data = self.service.get('/networks')
        return data['networks']

    @property
    def _external_services(self):
        CACHE_VALIDITY = 1  # seconds
        if self._last_got_external_services + CACHE_VALIDITY < time():
            self._last_got_external_services = time()
            self.log.debug('Fetching external services')
            try:
                data = self.service.get('/external-services/')
            except RequestException as e:
                self.log.warning('Cluster Config Service not available: %s', e)
                return
            services = data['external_services']
            self._last_external_services = []
            for service in services:
                hostname = service['external_host']
                port = service['external_port']
                try:
                    ip = get_host_ip(hostname, port)
                    if not ip:
                        raise Exception('No IPv6 address found for {}'.format(hostname))
                    service['ipv6'] = ip
                    self._last_external_services.append(service)
                except Exception as e:
                    self.log.warning('Failed to resolve IPv6 address of %s for service %s:%s: %s',
                                     hostname, service['service_name'], service['service_version'], e)
        else:
            self.log.debug('Using Cached External Services')
        return self._last_external_services

    def discover_external_service(self, zone_name: str, service_name: str, service_version: str) -> list:
        return [ServiceEndpoint(s['ipv6'], s['external_port']) for s in self._external_services
                if s['service_name'] == service_name
                and version_matches(s['service_version'], service_version)
                and s['zone_name'] == zone_name]

    def get_credentials(self, name: str) -> dict:
        data = self.service.get('/nodes/{}/credentials'.format(name))
        return data

    def get_zone(self, name: str) -> Zone:
        data = self.service.get('/zones/{}'.format(name))
        return Zone(data)


class ClusterVaults(ClusterCore):

    def get_vaults(self, name, zone):
        url = '/vaults/{vault_name}/{zone_name}'.format(vault_name=name, zone_name=zone)
        data = self.service.get(url, parse_json=False)
        return BytesIO(data)

    def extract_vault(self, name, zone, path):
        tar = tarfile.open(fileobj=self.get_vaults(name, zone))
        # TODO sanity checks
        tar.extractall(str(path))