from asyncio import get_event_loop
from logging import getLogger
import pequod_agent
from pequod_agent.models.node_state import NodeState
from pequod_agent.networking.ip import format_ip
import json
import requests
import random

random.seed()


class ClusterService:

    def __init__(self, credentials):
        self.log = getLogger('pequod.cluster-service')
        self.loop = get_event_loop()
        self.credentials = credentials
        self.user_agent = 'pequod-agent/{} {}-node/{}'.format(pequod_agent.__version__,
                                                              NodeState.get_template_name(),
                                                              NodeState.get_template_version())
        self.last_response_headers = None

    @property
    def endpoints(self) -> list:
        return []

    @property
    def name(self) -> str:
        raise NotImplementedError()

    def _request(self, method_name, method, path, use_credentials=True, *args, **kwargs):
        self._set_timeout(kwargs)
        if use_credentials:
            self._set_credentials(kwargs)
        else:
            kwargs['auth'] = None

        parse_json = kwargs.get('parse_json', True)
        try:
            del kwargs['parse_json']
        except KeyError:
            pass  # if it doesn't exist great

        headers = kwargs.setdefault('headers', {})
        headers['User-Agent'] = self.user_agent

        endpoints = self.endpoints
        if not endpoints:
            raise requests.ConnectionError('No endpoints found for {}'.format(self.name))
        random.shuffle(endpoints)
        last_error = None
        for endpoint in endpoints:
            url = 'http://{ip}:{port}{path}'.format(ip=format_ip(endpoint.ip), port=endpoint.port, path=path)
            try:
                response = method(url, *args, **kwargs)
            except requests.ConnectionError as e:
                last_error = 'Connection error for {}: {}'.format(url, e)
                self.log.warning(last_error)
                continue
            self.last_response_headers = response.headers
            if response.ok:
                self.log.debug('%s %s', method_name, response.url)
                try:
                    return response.json() if parse_json else response.content
                except ValueError:
                    return None
            elif response.status_code == 401 and use_credentials:
                self.log.debug('Got status 401 on %s. Retrying request without credentials..', response.url)
                # we need to pass parse_json status
                kwargs['parse_json'] = parse_json
                return self._request(method_name, method, path, False, *args, **kwargs)
            elif 500 <= response.status_code <= 599:  # Server errors: Retry
                last_error = 'Error {} on {}: {}'.format(response.status_code, response.url, response.text)
                self.log.warning(last_error)
            else:  # Client Error: Raise
                response.raise_for_status()
        error_message = 'Request to {} failed for all {} endpoint(s): {}'.format(self.name, len(endpoints), last_error)
        raise requests.ConnectionError(error_message)

    def _prepare_data(self, kwargs: dict):
        """
        Prepares data to be sent to server (on POST, PATCH and PUT). By default serializes the data in JSON.
        Kwargs is modified in place.
        """
        if 'headers' not in kwargs:
            # By default send data as json but still leave the possibility to use something else downstream
            kwargs['headers'] = {'Content-Type': 'application/json'}
            kwargs['data'] = json.dumps(kwargs.get('data', {}))

    def _set_timeout(self, kwargs: dict):
        """
        Sets a default timeout if it's not set already
        """
        if 'timeout' not in kwargs:
            kwargs['timeout'] = 10

    def _set_credentials(self, kwargs: dict):
        if self.credentials:
            kwargs['auth'] = (self.credentials['username'], self.credentials['password'])

    def delete(self, path, *args, **kwargs):
        # NOTE: we also need the Content-Type header for the controller's DELETE operation
        self._prepare_data(kwargs)
        return self._request('DELETE', requests.delete, path, *args, **kwargs)

    def get(self, path, *args, **kwargs):
        return self._request('GET', requests.get, path, *args, **kwargs)

    def patch(self, path, *args, **kwargs):
        self._prepare_data(kwargs)
        return self._request('PATCH', requests.patch, path, *args, **kwargs)

    def post(self, path, *args, **kwargs):
        self._prepare_data(kwargs)
        return self._request('POST', requests.post, path, *args, **kwargs)

    def put(self, path, *args, **kwargs):
        self._prepare_data(kwargs)
        return self._request('PUT', requests.put, path, *args, **kwargs)