#!/usr/bin/env python3

import subprocess
import time
import socket


class AddressConfigurationTimeout(Exception):
    def __init__(self, msg):
        Exception.__init__(self, msg)


def format_ip(ip: str) -> str:
    """
    IPv6 IPs need to be enclosed for URLs in brackets
    """
    if ':' in ip:
        return '[{}]'.format(ip)
    return ip


def wait_for_dad(dev: str, max_wait=30):
    """
    Wait for IPv6 Duplicate Address Detection (DAD) to complete on a given device.
    """
    start = time.time()
    while True:
        out = subprocess.check_output(['ip', '-o', 'a', 's', 'dev', dev, 'scope', 'global', 'tentative'])
        if not out:
            # OK, no more "tentative" addresses on device
            return
        if time.time() - start > max_wait:
            raise AddressConfigurationTimeout('IPv6 address configuration took longer than {}s on {}'.format(max_wait,
                                                                                                             dev))
        time.sleep(0.1)


def get_host_ip(host: str, port=None):
    try:
        socket.inet_pton(socket.AF_INET6, host)
        # if this didn't raise an exception it's already and valid ip
        return host
    except OSError:
        addrinfo = socket.getaddrinfo(host, port)
        for address in addrinfo:
            family, _, _, _, ip = address
            if family == socket.AF_INET6:
                return ip[0]

if __name__ == '__main__':
    subprocess.check_call(['ip', 'a', 'a', 'fd12:34::14/64', 'dev', 'eth0'])
    wait_for_dad('eth0')
