#!/usr/bin/env python3

import logging
import os
import shutil
import netaddr
import netifaces
import pequod_agent.drydock.networking as net
from pequod_agent.constants.net import BRIDGE_NAME
from pequod_agent.constants.path import PATHS_TO_SETUP, RESOLV_CONF_PATH
from pequod_agent.errors import PreflightError
from pequod_agent.networking.ip import get_host_ip, wait_for_dad
from pequod_agent.networking.ippool import IPPool
from pequod_agent.utilities.system import execute

log = logging.getLogger('pequod.setup')


def setup_agent(networks: list):
    make_dirs()
    ip_pool = IPPool.autoconfigure()
    setup_bridge(ip_pool)
    agent_network = setup_routes(ip_pool.net, networks)
    return ip_pool, agent_network


def make_dirs():
    # TODO test this
    log.debug('Creating directories')
    for directory in PATHS_TO_SETUP:
            try:
                directory.mkdir(parents=True)
            except FileExistsError:
                # the dir already exists great
                pass


def setup_bridge(ip_pool: IPPool):
    # TODO test this
    out = execute('brctl show {bridge}', False, bridge=BRIDGE_NAME)
    if 'No such device' in out:
        gateways = netifaces.gateways()['default']
        execute('brctl addbr {bridge}', bridge=BRIDGE_NAME)
        execute('ip link set dev {bridge} up', bridge=BRIDGE_NAME)
        iface = ip_pool.iface
        execute('brctl addif {bridge} {interface}', bridge=BRIDGE_NAME, interface=iface)

        addresses = netifaces.ifaddresses(iface)
        inet = addresses.get(netifaces.AF_INET)
        inet6 = addresses.get(netifaces.AF_INET6)
        for a in inet + inet6:
            if '%' not in a['addr']:
                net = netaddr.IPNetwork('{}/{}'.format(a['addr'], a['netmask']))
                execute('ip addr add {network} dev {bridge}', network=net, bridge=BRIDGE_NAME)
                execute('ip addr del {network} dev {device}', network=net, device=iface)
        # wait for IPv6 duplicate address detection
        wait_for_dad(BRIDGE_NAME)
        # add default gateways back to bridge device:
        gw = gateways.get(netifaces.AF_INET)
        if gw and gw[1] == iface:
            execute('ip route add default via {nexthop} dev {bridge}', nexthop=gw[0], bridge=BRIDGE_NAME)
        gw = gateways.get(netifaces.AF_INET6)
        if gw and gw[1] == iface:
            # note sure why, but we should remove the old default route
            execute('ip -6 route del default via {nexthop} dev {interface}', nexthop=gw[0], interface=iface)
            execute('ip -6 route add default via {nexthop} dev {bridge}', nexthop=gw[0], bridge=BRIDGE_NAME)
        else:
            log.warning('No IPv6 default gateway found: %s', gw)


def setup_routes(agent_network: netaddr.IPNetwork, networks: list) -> dict:
    # TODO test this
    log.info('Getting locations to setup routes..')

    log.debug('My IPv6 network is %s', agent_network)
    agent_location = find_my_network(networks, agent_network)
    if agent_location:
        log.info('Location is %s, default gateway is %s, nameserver is %s',
                 agent_location['location_name'], agent_location['ipv6_gateway'], agent_location['ipv6_dns'])
        execute('ip -6 route add default via {ipv6_gateway}', **agent_location)
        net.setup_routes(agent_location['ipv6_routes'])
        with RESOLV_CONF_PATH.open('w', encoding='ascii') as fd:
            fd.write('nameserver {}\n'.format(agent_location['ipv6_dns']))
    else:
        log.warning('Could not find my own location by IPv6 subnet (%s)', agent_network)
    return agent_location


def find_my_network(networks: list, my_net: netaddr.IPNetwork) -> dict:
    for loc in networks:
        # find our own location
        network = netaddr.IPNetwork('{}/{}'.format(loc['ipv6_subnet_ip'], loc['ipv6_subnet_prefixlen']))
        if network == my_net:
            return loc


def check_environment():
    # TODO: test this
    hostname = 'zalando.de'
    try:
        get_host_ip(hostname)
    except:
        log.warning('Could not resolve IPv6 address of %s. DNS/DNS64 seems not to work.', hostname)


def check(msg: str, assertion: bool):
    """
    Helper function to raise a PreflightError if the assertion failed

    >>> check('ok', True)
    >>> check('error', False)
    Traceback (most recent call last):
      ...
    pequod_agent.errors.PreflightError: Preflight check failed: error
    """
    if not assertion:
        raise PreflightError(msg)


def preflight_checks():
    """
    Check that all requirements are met before doing anything real
    """
    check('The agent must be run as root', os.getuid() == 0)
    check('Bridge utils must be installed', shutil.which('brctl'))
    check('HAProxy must be installed', shutil.which('haproxy'))