import logging
import sys


class ColorLoggingHandler(logging.StreamHandler):
    """
    Shamelessly stolen from
    https://github.com/laysakura/rainbow_logging_handler/blob/master/rainbow_logging_handler/__init__.py
    """

    color_map = {
        'black': 0,
        'red': 1,
        'green': 2,
        'yellow': 3,
        'blue': 4,
        'magenta': 5,
        'cyan': 6,
        'white': 7,
    }
    (csi, reset) = ('\x1b[', '\x1b[0m')

    # : Color of each column
    _column_color = {
        '%(asctime)s': ('black', None, True),
        '%(levelname)-7s': ('white', None, True),
        '%(name)s': ('white', None, False),
        '%(message)s': {
            logging.DEBUG: ('cyan', None, False),
            logging.INFO: ('white', None, True),
            logging.WARNING: ('yellow', None, True),
            logging.ERROR: ('red', None, True),
        },
    }

    def __init__(self, stream, fmt='%(asctime)s %(levelname)-7s %(name)s: %(message)s', datefmt='%Y-%m-%d %H:%M:%S'):
        super().__init__(stream)
        self._fmt = fmt
        self._datefmt = datefmt

        logging.StreamHandler.setFormatter(self, logging.Formatter(fmt=fmt, datefmt=datefmt))

    @property
    def is_tty(self):
        """Returns true if the handler's stream is a terminal."""
        return getattr(self.stream, 'isatty', lambda: False)()

    def get_color(self, fg=None, bg=None, bold=False):
        """
        Construct a terminal color code

        :param fg: Symbolic name of foreground color

        :param bg: Symbolic name of background color

        :param bold: Brightness bit
        """
        params = []
        if bg in self.color_map:
            params.append(str(self.color_map[bg] + 40))
        if fg in self.color_map:
            params.append(str(self.color_map[fg] + 30))
        if bold:
            params.append('1')

        color_code = ''.join((self.csi, ';'.join(params), 'm'))

        return color_code

    def emit(self, record):
        """Emit colorized `record` when called from `logging` module's printing functions"""
        try:
            msg = self.format(record)
            self.stream.write(msg + '\n')
            self.flush()
        except (KeyboardInterrupt, SystemExit):
            raise
        except:
            self.handleError(record)

    def colorize_traceback(self, formatter, record):
        """
        Turn traceback text to red.
        """
        if record.exc_info:
            # Cache the traceback text to avoid converting it multiple times
            # (it's constant anyway)
            record.exc_text = "".join([
                self.get_color("red"),
                formatter.formatException(record.exc_info),
                self.reset,
            ])

    def colorize_message(self, record):
        if 'Successful' in record.msg:
            record.msg = ''.join([self.get_color('green', bold=True), record.msg, self.reset])

    def colorize(self, record):
        """
        Get a special format string with ASCII color codes.
        """
        color_fmt = self._colorize_fmt(self._fmt, record.levelno)
        formatter = logging.Formatter(color_fmt, self._datefmt)
        self.colorize_traceback(formatter, record)
        self.colorize_message(record)
        output = formatter.format(record)
        # Clean cache so the color codes of traceback don't leak to other formatters
        record.ext_text = None
        return output

    def _colorize_fmt(self, fmt, levelno):
        """Adds ANSI color codes on plain `fmt`"""
        for column in self._column_color.keys():
            pos = fmt.find(column)
            if pos == -1:
                continue
            (pre_col, post_col) = (fmt[:pos], fmt[pos + len(column):])
            color_tup = self._column_color[column] if column != '%(message)s' else self._column_color[column].get(
                levelno, ('yellow', 'black', True))
            fmt = ''.join([pre_col,
                           self.reset, self.get_color(*color_tup), column, self.reset,
                           post_col])
        return fmt

    def format(self, record):
        """
        Formats a record for output.

        Takes a custom formatting path on a terminal.
        """
        if self.is_tty:
            message = self.colorize(record)
        else:
            message = logging.StreamHandler.format(self, record)

        return message


if __name__ == '__main__':
    logger = logging.getLogger()
    logger.setLevel(logging.DEBUG)
    handler = ColorLoggingHandler(sys.stderr)
    logger.addHandler(handler)
    logger.info("hello custom world")
    logging.warning('blub')
