#!/usr/bin/env python3
import json
from functools import wraps
from pathlib import Path


def coerce(f):
    annotation = f.__annotations__
    parameters = f.__code__.co_varnames

    def wrapper(*args, **kwargs):
        # add positional arguments to keyword argument dict
        for i, value in enumerate(args):
            name = parameters[i]
            kwargs[name] = value

        coerced_kwargs = dict()
        for name, value in kwargs.items():
            annotated_type = annotation.get(name, None)
            if annotated_type is not None:
                try:
                    value = annotated_type() if value is None else annotated_type(value)
                except TypeError:
                    # If the value can't be coerced leave the value as it is
                    pass
            coerced_kwargs[name] = value

        return f(**coerced_kwargs)
    return wrapper


def cached_json(cache_path: Path, filename: str):
    """
    decorator to cache function results as JSON

    * function arguments MUST be strings
    * function result MUST be JSON serializable
    """

    def decorating_function(func):

        @wraps(func)
        def wrapper(*args, **kwargs) -> dict:
            path = cache_path
            for arg in args:
                if isinstance(arg, str):
                    path /= arg
            path /= filename
            if path.exists():
                with path.open() as fd:
                    return json.load(fd)
            data = func(*args, **kwargs)
            if data:
                try:
                    path.parent.mkdir(parents=True)
                except FileExistsError:
                    pass
                with path.open('w') as fd:
                    json.dump(data, fd)
            return data
        return wrapper
    return decorating_function